"""
Database schema for collecting things
"""

from __future__ import annotations

from sqlalchemy import ForeignKey, String, Table, Column, Integer, event
from sqlalchemy.orm import Mapped, mapped_column, relationship

from moat.db.schema import Base


boxtyp_tree = Table(
    "boxtyp_tree",
    Base.metadata,
    Column(
        "parent_id",
        Integer,
        ForeignKey("boxtyp.id", name="fk_boxtyp_parent"),
        primary_key=True,
    ),
    Column("child_id", Integer, ForeignKey("boxtyp.id", name="fk_boxtyp_child"), primary_key=True),
)


class BoxTyp(Base):
    "One kind of box."

    name: Mapped[str] = mapped_column(unique=True, type_=String(40))
    comment: Mapped[str] = mapped_column(type_=String(200), nullable=True)

    parents: Mapped[set[BoxTyp]] = relationship(
        "BoxTyp",
        secondary=boxtyp_tree,
        primaryjoin="BoxTyp.id == boxtyp_tree.c.child_id",
        secondaryjoin="BoxTyp.id == boxtyp_tree.c.parent_id",
        back_populates="children",
    )
    children: Mapped[set[BoxTyp]] = relationship(
        "BoxTyp",
        secondary=boxtyp_tree,
        primaryjoin="BoxTyp.id == boxtyp_tree.c.parent_id",
        secondaryjoin="BoxTyp.id == boxtyp_tree.c.child_id",
        back_populates="parents",
    )
    usable: Mapped[bool] = mapped_column(
        nullable=False,
        default=True,
        server_default="1",
        comment="Can you put things directly into this?",
    )

    # Possible locations in there
    pos_x: Mapped[int] = mapped_column(nullable=True, comment="Max # of X positions")
    pos_y: Mapped[int] = mapped_column(nullable=True, comment="Max # of Y positions")
    pos_z: Mapped[int] = mapped_column(nullable=True, comment="Max # of Z positions")

    labeltyp_id: Mapped[int] = mapped_column(
        ForeignKey("labeltyp.id", name="fk_boxtyp_labeltyp"),
        nullable=True,
        comment="Default label",
    )

    boxes: Mapped[set[Box]] = relationship(back_populates="boxtyp")

    def dump(self):
        res = super().dump()
        res.pop("pos_x", None)
        res.pop("pos_y", None)
        res.pop("pos_z", None)
        if self.pos_x or self.pos_y or self.pos_z:
            res["size"] = [self.pos_x, self.pos_y, self.pos_z]
        if self.parents:
            res["parent"] = [p.name for p in self.parents]
        if self.children:
            res["child"] = [p.name for p in self.children]
        return res


class Box(Base):
    "One particular box, possibly with other boxes inside."

    id: Mapped[int] = mapped_column(primary_key=True)

    typ_id: Mapped[int] = mapped_column(ForeignKey("boxtyp.id", name="fk_box_typ"))
    name: Mapped[str] = mapped_column(unique=True, type_=String(40))
    container_id: Mapped[int] = mapped_column(
        ForeignKey("box.id", name="fk_box_container"),
        nullable=True,
    )

    boxtyp: Mapped[BoxTyp] = relationship(back_populates="boxes")
    container: Mapped[Box | None] = relationship(back_populates="boxes", remote_side=[id])
    boxes: Mapped[set[Box]] = relationship(back_populates="container")

    # location within its parent
    pos_x: Mapped[int] = mapped_column(nullable=True, comment="X position in parent")
    pos_y: Mapped[int] = mapped_column(nullable=True, comment="Y position in parent")
    pos_z: Mapped[int] = mapped_column(nullable=True, comment="Z position in parent")

    def dump(self):
        res = super().dump()
        res.pop("pos_x", None)
        res.pop("pos_y", None)
        res.pop("pos_z", None)
        if self.pos_x or self.pos_y or self.pos_z:
            res["pos"] = [self.pos_x, self.pos_y, self.pos_z]
        if self.container is not None:
            res["in"] = self.container.name
        if self.labels:
            res["labels"] = [f"{lab.id}:{lab.text}" for lab in self.labels]
        if self.boxes:
            res["content"] = sorted([box.name for box in self.boxes])
        if self.boxtyp:
            res["typ"] = self.boxtyp.name
        return res


@event.listens_for(Box, "before_insert")
@event.listens_for(Box, "before_update")
def validate_box_coords(mapper, connection, model):
    par = model.container
    if par is not None:
        par = par.boxtyp

    def chk(p):
        pp = f"pos_{p}"
        ppos = getattr(par, pp, None)

        if (pos := getattr(model, pp)) is None:
            if ppos is not None and ppos > 1:
                raise ValueError(f"Box {model.name} needs a value for {p}")
        elif pos <= 0:
            raise ValueError(f"Box {model.name} can't set {p} <= 0")
        elif par is None:
            raise ValueError(f"Box {model.name} is not in a sized container")
        elif ppos is None:
            raise ValueError(f"Box {model.name} can't have a value for {p}")
        elif ppos < pos:
            raise ValueError(f"Box {model.name}: {p} is {pos}, max is {ppos}")

    chk("x")
    chk("y")
    chk("z")



"""
Support for boxes for storing things.
"""

# The main code must not load any sqlalchemy code.
# sqlalchemy might not be present.

from __future__ import annotations

import sys
from moat.util import load_subgroup, ensure_cfg, yprint, option_ng
from moat.db import database
from .model import Box, BoxTyp
from sqlalchemy import select

import asyncclick as click

ensure_cfg("moat.db")


@load_subgroup(prefix="moat.box", invoke_without_command=False)
@click.pass_context
def cli(ctx):
    """Boxes for storing things."""
    obj = ctx.obj

    sess = ctx.with_resource(database(obj.cfg.db))
    ctx.with_resource(sess.begin())

    obj.session = sess


@cli.group
@click.option("--name", "-n", type=str, help="Text on the label")
@click.pass_obj
def one(obj, name):
    """
    Manage boxes.
    """
    obj.name = name
    pass


@one.command(name="show")
@click.pass_obj
def show_(obj):
    """
    Show box details / list all box types.

    If a name is set, show details for this box.
    Otherwise list all boxes without parents.
    """
    sess = obj.session

    if obj.name is None:
        seen = False
        with sess.execute(select(Box).where(Box.container == None)) as boxes:
            for (box,) in boxes:
                seen = True
                print(box.name)
        if not seen:
            print("No box types defined yet. Use '--help'?", file=sys.stderr)
    else:
        box = sess.one(Box, name=obj.name)
        yprint(box.dump())


def opts(c):
    c = option_ng("--name", "-n", type=str, help="Rename this box")(c)
    c = option_ng("--x", "-x", "pos_x", type=int, help="X position in parent")(c)
    c = option_ng("--y", "-y", "pos_y", type=int, help="Y position in parent")(c)
    c = option_ng("--z", "-z", "pos_z", type=int, help="Z position in parent")(c)
    c = option_ng("--in", "-i", "container", type=str, help="Where is it?")(c)
    c = option_ng("--typ", "-t", "boxtyp", type=str, help="Type of the box")(c)
    return c


@one.command()
@opts
@click.pass_obj
def add(obj, **kw):
    """
    Add a box.
    """
    if obj.name is None:
        raise click.UsageError("The box needs a name!")

    try:
        box = obj.session.one(Box, name=obj.name)
    except KeyError:
        pass
    else:
        print("This box already exists", file=sys.stderr)
        sys.exit(1)
    box = Box(name=obj.name)
    obj.session.add(box)
    box.apply(**kw)


@one.command(epilog="Use '-in -' to drop the parent box, -x/-y/-z 0 to clear a position.")
@opts
@click.pass_obj
def set(obj, **kw):
    """
    Modify a box.
    """
    if obj.name is None:
        raise click.UsageError("Which box? Use a name")

    try:
        box = obj.session.one(Box, name=obj.name)
    except KeyError:
        print("This box doesn't exist", file=sys.stderr)
        sys.exit(1)
    box.apply(**kw)


@one.command()
@click.pass_obj
def delete(obj):
    """
    Remove a box.
    """
    if obj.name is None:
        raise click.UsageError("Which box? Use a name")
    try:
        box = obj.session.one(Box, name=obj.name)
    except KeyError:
        print("This box doesn't exist", file=sys.stderr)
        sys.exit(1)
    obj.session.delete(box)


@cli.group(name="typ")
@click.option("--name", "-n", type=str, help="Name of the box type")
@click.pass_context
def typ_(ctx, name):
    """\
    Manage box types.
    """
    obj = ctx.obj
    obj.name = name


@typ_.command(name="show")
@click.pass_obj
def typ_show(obj):
    """
    Show box details / list all box types.

    If a name is set, show details for this type.
    Otherwise list all known box types.
    """
    sess = obj.session

    if obj.name is None:
        seen = False
        with sess.execute(select(BoxTyp)) as boxes:
            for (box,) in boxes:
                seen = True
                print(box.name)
        if not seen:
            print("No box types defined yet. Use '--help'?", file=sys.stderr)
    else:
        box = sess.one(BoxTyp, name=obj.name)
        yprint(box.dump())


def typopts(c):
    c = option_ng("--name", "-n", type=str, help="Rename this type")(c)
    c = option_ng("--x", "-x", "pos_x", type=int, help="Number of X positions for content")(c)
    c = option_ng("--y", "-y", "pos_y", type=int, help="Number of Y positions for content")(c)
    c = option_ng("--z", "-z", "pos_z", type=int, help="Number of Z positions for content")(c)
    c = option_ng("--comment", "-c", "comment", type=str, help="Description of this type")(c)
    c = click.option(
        "--in",
        "-i",
        "parent",
        type=str,
        multiple=True,
        help="Where can you put this thing into?",
    )(c)
    c = click.option("--usable", "-u", is_flag=True, help="Things can be put into this box")(c)
    c = click.option("--unusable", "-U", is_flag=True, help="The box holds fixed subdividers")(c)
    return c


@typ_.command(name="add")
@typopts
@click.pass_obj
def typ_add(obj, parent, **kw):
    """
    Add a box type.
    """
    if obj.name is None:
        raise click.UsageError("The box type needs a name!")

    bt = BoxTyp(name=obj.name)
    obj.session.add(bt)
    bt.apply(**kw)


@typ_.command(name="set", epilog="Use '-i -NAME' to remove a containing box type.")
@typopts
@click.pass_obj
def typ_set(obj, **kw):
    """
    Modify a box type.
    """
    if obj.name is None:
        raise click.UsageError("Which box type? Use a name")

    bt = obj.session.one(BoxTyp, name=obj.name)
    bt.apply(**kw)


@typ_.command(name="delete")
@click.pass_obj
def typ_delete(obj):
    """
    Remove a box type.
    """
    if obj.name is None:
        raise click.UsageError("Which box type? Use a name")
    bt = obj.session.one(BoxTyp, name=obj.name)
    obj.session.delete(bt)



"""
Database schema for collecting things
"""

from __future__ import annotations

from sqlalchemy.orm import relationship

from moat.db.schema import Base
from moat.db.util import session
from moat.util import NotGiven


from .model import Box, BoxTyp
from moat.label.model import Label, LabelTyp
from moat.thing.model import Thing
from moat.db.schema import Base

BoxTyp.labeltyp = relationship(LabelTyp, back_populates="boxtypes")
Box.labels = relationship(Label, back_populates="box", collection_class=set)
Box.things = relationship(Thing, back_populates="container", collection_class=set)


def box_apply(self, container=NotGiven, boxtyp=NotGiven, **kw):
    sess = session.get()
    with sess.no_autoflush:
        Base.apply(self, **kw)

        if container is NotGiven:
            pass
        elif container is None:
            self.container = None
        else:
            self.container = sess.one(Box, name=container)

        if boxtyp is NotGiven:
            if self.boxtyp is None:
                raise ValueError("New boxes need a type")
        elif boxtyp is None:
            raise ValueError("Boxes need a type")
        else:
            if self.boxtyp is None:
                self.boxtyp = sess.one(BoxTyp, name=boxtyp)
            elif self.boxtyp.name != boxtyp:
                raise ValueError("Box types cannot be changed")


Box.apply = box_apply


def boxtyp_apply(self, usable, unusable, parent=(), **kw):
    Base.apply(self, **kw)
    if usable:
        if unusable:
            raise ValueError("Can't both use and not use this type")
        self.usable = True
    elif unusable:
        self.usable = False

    for p in parent:
        if p[0] != "-":
            self.parents.add(session.get().one(BoxTyp, name=p))
        else:
            self.parents.remove(session.get().one(BoxTyp, name=p[1:]))


BoxTyp.apply = boxtyp_apply
